// Code generated by "stringer -type=MarkupKind -output=markup_kind_string.go"; DO NOT EDIT.

package lang

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[NilKind-0]
	_ = x[PlainTextKind-1]
	_ = x[MarkdownKind-2]
}

const _MarkupKind_name = "NilKindPlainTextKindMarkdownKind"

var _MarkupKind_index = [...]uint8{0, 7, 20, 32}

func (i MarkupKind) String() string {
	if i >= MarkupKind(len(_MarkupKind_index)-1) {
		return "MarkupKind(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _MarkupKind_name[_MarkupKind_index[i]:_MarkupKind_index[i+1]]
}
